"""General utilities module."""
import asyncio
import contextvars
import functools
from typing import Callable
from typing import TypeVar

from typing_extensions import ParamSpec


_P = ParamSpec("_P")
_R = TypeVar("_R")


async def to_thread(func: Callable[_P, _R], *args: _P.args, **kwargs: _P.kwargs) -> _R:
    """Backport of asyncio.to_thread."""
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)  # type: ignore[arg-type]



"""CLI progress module."""
from typing import Optional
from typing import Union

import tqdm.std


class ProgressCallback:
    """Callbacks for displaying progress in CLI commands."""

    def __init__(self, pbar: tqdm.std.tqdm) -> None:
        """Construct a callback."""
        self.pbar = pbar

    def set_desc(self, desc: str) -> None:
        """Set pbar description."""
        self.pbar.set_description(desc)

    def set_total(self, total: Optional[Union[int, float]]) -> None:
        """Set pbar total."""
        if total is not None:  # pragma: no cover
            self.pbar.total = total
            self.pbar.refresh()

    def update(self, inc: Union[int, float] = 1) -> None:
        """Increment pbar progress."""
        self.pbar.update(inc)



"""LPEG API credentials."""
import hashlib
import os
import platform
import uuid
from dataclasses import dataclass
from typing import ClassVar
from typing import Dict
from typing import Union
from typing import cast

from ..exceptions import NoCredentialsError


if platform.system().lower() == "windows":
    import winreg
else:
    import fake_winreg as winreg  # type: ignore


@dataclass
class BaseCredentials:
    """Base LPEG API credentials.

    Attributes:
        uid: LPEG API user ID (login username).
        st: LPEG API st token (6-digit numeric string).
        mid: LPEG API mid (64-character alphanumeric string).
        pid: LPEG API player ID (alphanumeric player ID).
    """

    uid: str
    st: str
    mid: str
    pid: str

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseCredentials):
            return False
        return (
            self.uid == other.uid
            and self.st == other.st
            and self.mid == other.mid
            and self.pid == other.pid
        )

    @classmethod
    def get_default(cls) -> "BaseCredentials":  # pragma: no cover
        """Lookup and return default credentials for the current user.

        Raises:
            NoCredentialsError: No default credentials for the current user
                could be found.
        """
        raise NoCredentialsError

    @classmethod
    def get_device_id(cls) -> "str":
        """Return a device ID suitable for generating a player ID (`pid`).

        LPEG player ID's consist of a a server token prepended to a unique
        device ID. Device ID generation is dependent on platform and client,
        but is comparable to ``UnityEngine.SystemInfo.deviceUniqueIdentifier``.

        Returns:
            Unique (hardware based) device ID roughly equivalent to an LPEG
            desktop player device ID.
        """
        hardware_id = uuid.getnode().to_bytes(length=6, byteorder="big")
        return hashlib.blake2b(hardware_id, digest_size=6).digest().hex()


@dataclass
class FourDCredentials(BaseCredentials):
    """4D Media Player API credentials.

    Note:
        Local credential lookup requires a registered 4D Media Player
        installation, meaning that the user must install the player, log in
        with a valid LPEG/Afesta account, and register that player instance
        to the user's account.

        When looking up local player credentials, afesta_tools will not read
        the local user's LPEG/Afesta password, it only reads the API access
        credentials which are already stored on the local machine by 4D Media
        Player.
    """

    WINREG_KEY: ClassVar[str] = r"SOFTWARE\lpeg\4D MEDIA PLAYER"
    PID_CONFIG_PATH: ClassVar[str] = os.path.join(
        "~",
        "AppData",
        "LocalLow",
        "lpeg",
        "4D MEDIA PLAYER",
        "PidConfiguration.json",
    )

    @classmethod
    def get_default(cls) -> "FourDCredentials":
        """Lookup and return default credentials for the current user.

        Returns:
            Default 4D Media Player credentials.

        Raises:
            NoCredentialsError: No default credentials for the current user
                could be found.
        """
        reg_values = cls._get_user_reg_values()
        return cls(
            uid=reg_values["login_account"],
            st=reg_values["st"],
            mid=reg_values["mid"],
            pid=cls._get_user_pid(),
        )

    @classmethod
    def _get_user_reg_values(cls) -> Dict[str, str]:
        result: Dict[str, str] = {}
        try:
            with winreg.OpenKey(  # type: ignore[attr-defined]
                winreg.HKEY_CURRENT_USER,  # type: ignore[attr-defined]
                cls.WINREG_KEY,
            ) as key:
                names = {"login_account", "mid", "st"}
                i = 0
                while True:
                    try:
                        enum_val = winreg.EnumValue(  # type: ignore[attr-defined]
                            key, i
                        )
                        value_name, value, typ = enum_val
                        for name in names:
                            if value_name.startswith(f"{name}_"):
                                result[name] = cls._coerce_reg_str(value, typ)
                    except OSError:
                        break
                    i += 1
        except (OSError, ValueError) as exc:
            raise NoCredentialsError(
                "No registered 4D Media Player installation could be found."
            ) from exc
        return result

    @staticmethod
    def _coerce_reg_str(value: Union[str, bytes], typ: int) -> str:
        if typ == winreg.REG_SZ:  # type: ignore[attr-defined]
            return cast(str, value).rstrip("\0")
        elif typ == winreg.REG_BINARY:  # type: ignore[attr-defined] # pragma: no cover
            try:
                return cast(bytes, value).decode("utf-8")
            except UnicodeDecodeError:
                pass
        raise ValueError(  # pragma: no cover
            f"Unexpected winreg value: {value!r}, {typ}"
        )

    @classmethod
    def _get_user_pid(cls) -> str:
        import json

        path = os.path.expanduser(cls.PID_CONFIG_PATH)
        try:
            with open(path, encoding="utf-8") as fp:
                pid_config: Dict[str, str] = json.load(fp)
            return pid_config["pid"]
        except (OSError, json.JSONDecodeError, KeyError) as exc:
            raise NoCredentialsError(
                "No registered 4D Media Player installation could be found."
            ) from exc



"""Command-line interface."""
import asyncio
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import cast

import click
from tqdm.asyncio import tqdm

from .config import dump_credentials
from .config import load_credentials
from .exceptions import AfestaError
from .exceptions import NoCredentialsError
from .lpeg.client import BaseLpegClient
from .lpeg.client import FourDClient
from .lpeg.credentials import BaseCredentials
from .lpeg.credentials import FourDCredentials
from .progress import ProgressCallback


@click.group()
@click.version_option()
def cli() -> None:
    """Afesta Tools."""


def _load_credentials() -> BaseCredentials:
    """Try to load default credentials.

    Will attempt to load creds from afesta-tools config or an existing
    4D Media Player installation, in that order.

    Returns:
        Default credentials.
    """
    try:
        return load_credentials()
    except NoCredentialsError:
        pass
    return FourDCredentials.get_default()


@cli.command()
@click.option("-u", "--username", default=None, help="Afesta username.")
@click.option("-p", "--password", default=None, help="Afesta password.")
@click.option(
    "-f", "--force", is_flag=True, default=False, help="Overwrite existing credentials."
)
def login(
    username: Optional[str], password: Optional[str], force: bool
) -> int:  # noqa: DAR101
    """Login to Afesta and register afesta-tools as a new player.

    If username and/or password are not specified, they will be prompted via the
    command-line.

    Login is not required if 4D Media Player is installed and the current
    user has logged into 4D Media Player and registered it with an Afesta
    account.

    Note that afesta-tools only stores username and API tokens (password will
    not be saved to disk).
    """
    if not force:
        try:
            creds = load_credentials()
            click.echo(f"Already logged in as {creds.uid}")
            return 0
        except NoCredentialsError:
            pass
    if not username:
        username = click.prompt("Afesta username")
    if not password:
        password = click.prompt("Afesta password", hide_input=True)
    try:
        username = cast(str, username)
        password = cast(str, password)
        creds = asyncio.run(_login(username, password))
        dump_credentials(creds)
        click.echo(f"Logged into Afesta as {creds.uid}")
    except AfestaError as exc:  # pragma: no cover
        click.echo(f"Login failed: {exc}", err=True)
        return 1
    return 0


async def _login(username: str, password: str) -> BaseCredentials:
    async with FourDClient() as client:
        return await client.register_player(username, password)


@cli.command()
@click.argument("video_id", nargs=-1)
def dl(video_id: Sequence[str]) -> int:  # noqa: DAR101
    """Download an afesta video.

    Requires an account with permissions to download the video (either via
    standalone purchase or monthly subscription DL benefits).

    If 4D Media Player is installed and the current user is logged in via the
    player, the existing 4D Media Player credentials will be used. Otherwise,
    the 'afesta login' command must be run before downloading.
    """
    try:
        creds = _load_credentials()
    except NoCredentialsError:
        click.echo("No credentials found. Did you forget to run 'afesta login'?")
    try:
        asyncio.run(_dl(video_id, creds))
    except AfestaError as exc:  # pragma: no cover
        click.echo(f"Download failed: {exc}", err=True)
        return 1
    return 0


async def _dl(video_ids: Sequence[str], creds: BaseCredentials) -> None:
    async with FourDClient(creds) as client:
        await asyncio.gather(*(_dl_one(client, video_id) for video_id in video_ids))


async def _dl_one(client: BaseLpegClient, video_id: str) -> None:
    with tqdm(unit="B", unit_scale=True) as pbar:
        await client.download_video(
            video_id,
            progress=ProgressCallback(pbar),
        )


@cli.command()
@click.argument("video_id", nargs=-1)
def dl_vcz(video_id: Sequence[str]) -> int:  # noqa: DAR101
    """Download vcz files for an afesta video.

    Requires an account with permissions to download the video (either via
    standalone purchase or monthly subscription DL benefits).

    If 4D Media Player is installed and the current user is logged in via the
    player, the existing 4D Media Player credentials will be used. Otherwise,
    the 'afesta login' command must be run before downloading.
    """
    try:
        creds = _load_credentials()
    except NoCredentialsError:
        click.echo("No credentials found. Did you forget to run 'afesta login'?")
    try:
        asyncio.run(_dl_vczs(video_id, creds))
    except AfestaError as exc:  # pragma: no cover
        click.echo(f"Download failed: {exc}", err=True)
        return 1
    return 0


async def _dl_vczs(video_ids: Sequence[str], creds: BaseCredentials) -> None:
    async with FourDClient(creds) as client:
        await asyncio.gather(*(_dl_vcz(client, video_id) for video_id in video_ids))


async def _dl_vcz(client: BaseLpegClient, video_id: str) -> None:
    with tqdm(unit="B", unit_scale=True) as pbar:
        await client.download_vcz(
            video_id,
            progress=ProgressCallback(pbar),
        )


@cli.command()
@click.argument(
    "filename",
    nargs=-1,
    type=click.Path(exists=True, dir_okay=False, readable=True),
)
@click.option(
    "--format",
    "fmt",
    type=click.Choice(["csv", "vcsx"], case_sensitive=False),
    default="csv",
    help="Script format (defaults to CSV).",
)
def extract_script(
    filename: Sequence[Path], fmt: Literal["csv", "vcsx"]
) -> int:  # noqa: DAR101
    """Extract interlocking goods script files from a VCZ file."""
    try:
        asyncio.run(_extract_script(filename, fmt))
    except AfestaError as exc:  # prgma: no cover
        click.echo(f"Script extraction failed: {exc}", err=True)
        return 1
    return 0


async def _extract_script(
    filenames: Sequence[Path], fmt: Literal["csv", "vcsx"]
) -> None:
    await asyncio.gather(*(_extract_one(name, fmt) for name in filenames))


async def _extract_one(filename: Path, fmt: Literal["csv", "vcsx"]) -> None:
    from .vcs import GoodsType
    from .vcs import VCZArchive

    async with VCZArchive(filename) as vcz:
        for typ in (
            GoodsType.CYCLONE,
            GoodsType.PISTON,
            GoodsType.ONARHYTHM,
        ):
            try:
                path = await vcz.extract_script(typ, fmt)
                click.echo(f"Extracted {path}")
            except (KeyError, ValueError):
                pass


if __name__ == "__main__":
    cli(prog_name="afesta")  # pragma: no cover



"""VCS interlocking goods module."""
import enum
import io
from typing import Literal
from typing import Union
from typing import cast

from a10sa_script.script import VCSXCycloneScript
from a10sa_script.script import VCSXOnaRhythmScript
from a10sa_script.script import VCSXPistonScript
from a10sa_script.script import VorzeLinearScript
from a10sa_script.script import VorzeRotateScript
from a10sa_script.script import VorzeVibrateScript


GoodsScript = Union[
    VCSXCycloneScript,
    VCSXOnaRhythmScript,
    VCSXPistonScript,
]
ConvertedScript = Union[
    GoodsScript,
    VorzeLinearScript,
    VorzeRotateScript,
    VorzeVibrateScript,
]
ScriptFormat = Literal["csv", "vcsx", "funscript"]


class GoodsType(enum.Enum):
    """Interlocking goods type.

    Attributes:
        CYCLONE: Vorze CycloneSA devices.
        PISTON: Vorze Piston devices.
        ONARHYTHM: Vorze OnaRhythm devices (Rocket+1D).
    """

    CYCLONE = "Vorze_CycloneSA"
    PISTON = "Vorze_Piston"
    ONARHYTHM = "Vorze_OnaRhythm"


def load_script(typ: GoodsType, data: bytes) -> GoodsScript:
    """Load interlocking goods script data.

    Arguments:
        typ: Goods type.
        data: Script binary (VCSX) data.

    Returns:
        VCSX script.

    Raises:
        ValueError: Invalid goods type.
    """
    with io.BytesIO(data) as f:
        if typ == GoodsType.CYCLONE:
            return cast(VCSXCycloneScript, VCSXCycloneScript.load(f))
        if typ == GoodsType.PISTON:
            return cast(VCSXPistonScript, VCSXPistonScript.load(f))
        if typ == GoodsType.ONARHYTHM:
            return cast(VCSXOnaRhythmScript, VCSXOnaRhythmScript.load(f))
    raise ValueError("Invalid goods type")


def convert_script(script: GoodsScript, fmt: ScriptFormat) -> ConvertedScript:
    """Convert interlocking goods script data.

    Arguments:
        script: Goods script.
        fmt: Output script format.

    Returns:
        Converted script.

    Raises:
        ValueError: Unsupported output format.
    """
    if fmt == "vcsx":
        return script
    if isinstance(script, VCSXCycloneScript):
        if fmt == "csv":
            return VorzeRotateScript(script.commands)
    if isinstance(script, VCSXOnaRhythmScript):
        if fmt == "csv":
            return VorzeVibrateScript(script.commands)
    if isinstance(script, VCSXPistonScript):
        if fmt == "csv":
            return VorzeLinearScript(script.commands)
    raise ValueError(f"Unable to convert {script} to format {fmt}")



"""Afesta Tools config module."""
import json
import os
from dataclasses import asdict

import platformdirs

from .exceptions import NoCredentialsError
from .lpeg.credentials import BaseCredentials


APP_NAME = "afesta-tools"
CREDENTIALS_FILE = "credentials.json"


def dump_credentials(creds: BaseCredentials) -> None:
    """Dump credentials to the default credentials file."""
    config_dir = platformdirs.user_config_dir(APP_NAME)
    os.makedirs(config_dir, exist_ok=True)
    creds_file = os.path.join(config_dir, CREDENTIALS_FILE)
    with open(creds_file, mode="w", encoding="utf-8") as f:
        json.dump(asdict(creds), f)


def load_credentials() -> BaseCredentials:
    """Load credentials from the default credentials file.

    Returns:
        Default credentials.

    Raises:
        NoCredentialsError: Credentials could not be loaded.
    """
    config_dir = platformdirs.user_config_dir(APP_NAME)
    os.makedirs(config_dir, exist_ok=True)
    creds_file = os.path.join(config_dir, CREDENTIALS_FILE)
    try:
        with open(creds_file, encoding="utf-8") as f:
            return BaseCredentials(**json.load(f))
    except (OSError, json.JSONDecodeError) as exc:
        raise NoCredentialsError("Failed to load default credentials.") from exc



"""VCZ (zip-compressed VCS) archive module."""
import os
import threading
import zipfile
from functools import cached_property
from typing import Any
from typing import AsyncContextManager
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from lxml import etree  # noqa: S410

from ..types import PathLike
from ..utils import to_thread
from .goods import GoodsScript
from .goods import GoodsType
from .goods import ScriptFormat
from .goods import convert_script
from .goods import load_script


class VCZArchive(AsyncContextManager["VCZArchive"]):
    """VCZ archive."""

    def __init__(self, filename: PathLike, mode: Literal["r", "a"] = "r") -> None:
        """Open a VCZ archive.

        Arguments:
            filename: Path to VCZ file.
            mode: ``r`` to read an existing file, ``a`` to append (edit) an existing
                file.

        Raises:
            ValueError: `filename` is not a valid VCZ archive.
        """
        self.filename = filename
        self._zip = zipfile.ZipFile(self.filename, mode=mode)
        self._lock = threading.Lock()
        params = self._load_params()
        sys_params = params.find("system")
        if sys_params is None:
            raise ValueError(f"{filename} has no VCS system params")
        self._sys_params = sys_params
        names = self._zip.namelist()
        self._name_infos = {
            element.tag: element.text
            for element in self._sys_params
            if element.text in names
        }

    def _load_params(self) -> etree._Element:
        """Return the contents of params.xml."""
        try:
            params_info = self._zip.getinfo("params.xml")
        except KeyError as e:
            raise ValueError(f"{self.filename} has no VCS params") from e
        parser = etree.XMLParser(resolve_entities=False)
        return etree.fromstring(  # noqa: S320
            self._zip_read(params_info), parser=parser
        )

    @cached_property
    def title(self) -> Optional[str]:
        """Video title."""
        return self._sys_params.findtext("title")

    def namelist(self) -> List[str]:
        """Return a list of archive members by name."""
        return list(self._name_infos.keys())

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.close()

    async def close(self) -> None:
        """Close this client."""
        self._zip.close()

    def _zip_read(self, name: Union[str, zipfile.ZipInfo]) -> bytes:
        with self._lock:
            return self._zip.read(name)

    async def read(self, name: str) -> bytes:
        """Read the specified file from this archive.

        Arguments:
            name: File to read.

        Returns:
            File data.

        Raises:
            KeyError: `name` does not exist in the archive.
        """
        return await to_thread(self._zip_read, self._name_infos[name])

    async def read_script(self, typ: GoodsType) -> GoodsScript:
        """Read the specified interlocking goods script.

        Arguments:
            typ: Goods type.

        Returns:
            VCSX script.

        Raises:
            KeyError: The specified goods type is not supported for this video.
        """
        data = await self.read(typ.value)
        return load_script(typ, data)

    async def extract_script(
        self,
        typ: GoodsType,
        fmt: ScriptFormat = "csv",
        path: Optional[PathLike] = None,
    ) -> str:
        """Extract the specified interlocking goods script.

        Arguments:
            typ: Goods type
            fmt: Output format for the extracted script. Defaults to ``csv``.
                Supported formats:
                    * ``csv`` - Vorze CSV
                    * ``vcsx`` - Afesta/LPEG binary vcsx
            path: Optional output directory in which to extract the script.
                Defaults to the current working directory.

        Returns:
            Path for the extracted script.

        Raises:
            KeyError: The specified goods type is not supported for this video.
            ValueError: The specified output format is not supported for this
                script type.
        """
        script = convert_script(await self.read_script(typ), fmt)
        filename = f"{self._script_stem(typ, fmt)}.{fmt}"
        if path:
            filename = os.path.join(path, filename)
        with open(filename, "wb") as f:
            script.dump(f)
        return filename

    def _script_stem(self, typ: GoodsType, fmt: ScriptFormat) -> str:
        base, _ = os.path.splitext(os.path.basename(self.filename))
        if fmt == "funscript":
            return base
        return {
            GoodsType.CYCLONE: f"{base}_cyclone",
            GoodsType.ONARHYTHM: f"{base}_onarhythm",
            GoodsType.PISTON: f"{base}_piston",
        }[typ]



"""LPEG API client."""
import enum
import os
from abc import abstractmethod
from contextlib import AsyncExitStack
from typing import Any
from typing import AsyncContextManager
from typing import Awaitable
from typing import Callable
from typing import Optional

import aiohttp
from funcy import wraps

from ..exceptions import AfestaError
from ..exceptions import AuthenticationError
from ..progress import ProgressCallback
from ..types import PathLike
from .credentials import BaseCredentials
from .credentials import FourDCredentials


AP_STATUS_CHK_URL = "https://www.lpeg.jp/manage/ap_status_chk.php"
AP_LOGIN_URL = "https://www.lpeg.jp/manage/ap_login.php"
AP_REG_URL = "http://www.lpeg.jp/manage/ap_reg.php"
DL_URL = "https://lpeg.jp/h/"
VCS_DL_URL = "https://data.lpeg.jp/ap_vcs_dl.php"


class VideoQuality(enum.Enum):
    """Video download quality.

    Attributes:
        H264: Best available H.264 quality.
        H265: Best avaialable HEVC quality.
        PC_SBS: Alias for H264.

    Note: `VideoQuality` maps qualities as they are listed in the LPEG web
        UI to the `type` field used in the backend LPEG API. The `type` string
        does not always correspond to the actual video codec and/or resolution.

        In general, for 4K VR content `PC_SBS` will map to PC 4K. In some
        cases, the `PC_SBS` video will be 3K, even though the video has a
        4K HEVC option available.
    """

    H264 = "h264"  # 3K/4K H264
    H265 = "h265"  # 3K/4K HEVC
    PC_SBS = "h264"  # PC 3K/4K


def require_auth(coroutine: Callable[..., Awaitable[Any]]) -> Any:
    """Decorator for API calls which require authentication.

    Arguments:
        coroutine: Coroutine to decorate.

    Returns:
        Decorated function.

    Raises:
        AuthenticationError: Auth credentials are unavailable.
    """

    @wraps(coroutine)  # type: ignore[misc]
    async def wrapper(obj: "BaseLpegClient", *args: Any, **kwargs: Any) -> Any:
        if not obj.creds:
            raise AuthenticationError(
                f"{coroutine.__name__} requires valid credentials."
            )
        return await coroutine(obj, *args, **kwargs)

    return wrapper


class BaseLpegClient(AsyncContextManager["BaseLpegClient"]):
    """lpeg.jp API client.

    Can be used as an async context manager. When used as a context manager,
    `close` will be called automatically on exit.
    """

    CHUNK_SIZE = 4096
    DEFAULT_VIDEO_QUALITY = VideoQuality.PC_SBS
    _CLIENT_TIMEOUT = 5 * 60

    def __init__(self, creds: Optional[BaseCredentials] = None) -> None:
        """Construct a new client.

        Arguments:
            creds: LPEG API credentials. Required to make authenticated API calls.
                Public (unauthenticated) API calls can still be made when `creds`
                is not set.
        """
        super().__init__()
        self.creds = creds
        self._exit_stack = AsyncExitStack()
        self._session = aiohttp.ClientSession(
            headers={"User-Agent": self.user_agent},
            raise_for_status=True,
        )

    async def __aenter__(self) -> "BaseLpegClient":
        await self._exit_stack.enter_async_context(self._session)
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.close()

    async def close(self) -> None:
        """Close this client."""
        async with self._exit_stack:
            pass

    @property
    @abstractmethod
    def user_agent(self) -> str:
        """Return the HTTP User-Agent for this client."""

    @property
    def _dl_timeout(self) -> aiohttp.ClientTimeout:
        """Return download timeout."""
        return aiohttp.ClientTimeout(
            total=None,
            connect=self._CLIENT_TIMEOUT,
            sock_connect=self._CLIENT_TIMEOUT,
            sock_read=None,
        )

    async def _get(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        return await self._session.get(url, **kwargs)

    async def _post(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        return await self._session.post(url, **kwargs)

    @require_auth
    async def status_chk(self) -> None:
        """Run ap_status_chk API request."""
        assert self.creds is not None
        payload = {
            "st": self.creds.st,
            "mid": self.creds.mid,
            "pid": self.creds.pid,
            "type": "dpvr",
        }
        await self._post(AP_STATUS_CHK_URL, data=payload)

    async def _download(
        self,
        response: aiohttp.ClientResponse,
        download_dir: Optional[PathLike] = None,
        progress: Optional[ProgressCallback] = None,
    ) -> None:
        if response.content_disposition:
            filename: Optional[str] = response.content_disposition.filename
        else:  # pragma: no cover
            filename = None
        if not filename:  # pragma: no cover
            raise AfestaError("Not a valid file download URL")
        if download_dir:
            filename = os.path.join(download_dir, filename)
        if progress:
            progress.set_desc(f"Downloading {filename}")
            if "Content-Length" in response.headers:  # pragma: no cover
                progress.set_total(int(response.headers["Content-Length"]))
        with open(filename, mode="wb") as fp:
            async for chunk in response.content.iter_chunked(self.CHUNK_SIZE):
                fp.write(chunk)
                if progress:
                    progress.update(len(chunk))

    @require_auth
    async def download_video(
        self,
        code: str,
        download_dir: Optional[PathLike] = None,
        quality: Optional[VideoQuality] = None,
        progress: Optional[ProgressCallback] = None,
    ) -> None:
        """Download a video.

        Arguments:
            code: Video code (LPEG tr... code).
            download_dir: Directory for downloaded files. Defaults to the
                current working dir.
            quality: Video quality. Defaults to `DEFAULT_VIDEO_QUALITY`.
            progress: Optional progress callback.

        Note:
            Actual download quality may be worse than the requested value
            depending on the video.
        """
        resp = await self._request_video(code, quality=quality)
        await self._download(resp, download_dir=download_dir, progress=progress)

    async def _request_video(
        self,
        code: str,
        quality: Optional[VideoQuality] = None,
    ) -> aiohttp.ClientResponse:
        assert self.creds is not None
        if quality is None:
            quality = self.DEFAULT_VIDEO_QUALITY
        params = {
            "op": 1,
            "type": quality.value,
            "code": code,
            "pid": self.creds.pid,
        }
        return await self._get(DL_URL, params=params, timeout=self._dl_timeout)

    async def download_vcz(
        self,
        fid: str,
        download_dir: Optional[PathLike] = None,
        progress: Optional["ProgressCallback"] = None,
    ) -> None:
        """Download a vcz.

        Arguments:
            fid: Video FID.
            download_dir: Directory for downloaded files. Defaults to the
                current working dir.
            progress: Optional progress callback.
        """
        await self.status_chk()
        resp = await self._request_vcz(fid)
        await self._download(resp, download_dir=download_dir, progress=progress)

    async def _request_vcz(self, fid: str) -> aiohttp.ClientResponse:
        assert self.creds is not None
        params = {
            "pid": self.creds.pid,
            "fid": fid,
        }
        headers = {"Accept-Encoding": "gzip, identity"}
        return await self._get(
            VCS_DL_URL, params=params, headers=headers, timeout=self._dl_timeout
        )

    async def register_player(
        self,
        username: str,
        password: str,
    ) -> BaseCredentials:
        """Login to LPEG and register as new VR-capable player.

        Arguments:
            username: Afesta/LPEG login username.
            password: Password for `username` account.

        Note:
            `password` will not be saved. Only `username` and API tokens are
            saved in credentials.

        Returns:
            Newly registered credentials.

        Raises:
            AuthenticationError: An authentication error occured.
        """
        device_id = BaseCredentials.get_device_id()
        try:
            resp = await self._get(AP_REG_URL, params={"pid": device_id})
            data = await resp.json()
            pid = data["mp_no"]
        except (aiohttp.ClientError, KeyError) as exc:
            raise AuthenticationError("Player registration failed.") from exc
        try:
            resp = await self._post(
                AP_LOGIN_URL,
                data={
                    "uid": username,
                    "pass": password,
                    "pid": pid,
                    "type": "dpvr",
                },
            )
            data = (await resp.json()).get("data", {})
            mid = data["mid"]
            st = data["st"]
        except (aiohttp.ClientError, KeyError) as exc:
            raise AuthenticationError("Login failed.") from exc
        self.creds = self.new_credentials(uid=username, st=st, mid=mid, pid=pid)
        return self.creds

    @abstractmethod
    def new_credentials(self, *args: Any, **kwargs: Any) -> BaseCredentials:
        """Return a new credentials instance."""


class FourDClient(BaseLpegClient):
    """4D Media Player client.

    All LPEG API calls will be made in the same way as 4D Media Player, except
    for video downloads. Video quality will default to `VideoQuality.PC_SBS`,
    which is the Afesta default in the web UI for PC downloads, and
    is the default for in-client 4D Media Player downloads.
    """

    @property
    def user_agent(self) -> str:
        """Return the HTTP User-Agent for this client."""
        return "BestHTTP 1.12.3"  # UA as of 4D Media Player 2.0.1

    def new_credentials(self, *args: Any, **kwargs: Any) -> FourDCredentials:
        """Return a new credentials instance."""
        return FourDCredentials(*args, **kwargs)



"""Afesta Tools exceptions."""


class AfestaError(Exception):
    """Base exception."""


class NoCredentialsError(AfestaError):
    """No valid credentials could be found."""


class AuthenticationError(AfestaError):
    """API authentication error."""
